@@ -19,7 +19,6 @@ class CsfRaster(Raster, GridEnvelope2D):
     __cellrepr = 'UNKNOWN'
     __valuescale = 'VS_UNDEFINED'
     __version = 2
-    __mode = 'r'
     dataformat = 'f'
     __MAIN_HDR = [
         {'type':'char', 'name':'signature', 'endpos':32},
@@ -221,9 +220,15 @@ class CsfRaster(Raster, GridEnvelope2D):
         
         # If file does not exist and mode[0] = 'w', create it!
         if (mode[0] == 'w') and (not self.file_exists):
+            # Go ahead and also assign _mode
             self.datafile = open(os.path.join(self.folder, self.name), 'wb')
-            self.__mode = mode;
-            GridEnvelope2D.__init__(self, ncols, nrows, xll, yll, cellsize, cellsize);
+            self._mode = mode;
+            
+            # If dx and dy have been set to different values already, make sure those are written to disk
+            if abs(self.dx - self.dy) < const.epsilon:
+                self.dx = cellsize
+                self.dy = cellsize
+            GridEnvelope2D.__init__(self, ncols, nrows, xll, yll, self.dx, self.dy);
             self.cellsize = cellsize;
             self.nodatavalue = nodatavalue;
             self.__set_header_value(self.__RASTER_HDR, "valueScale", valueScale)
