@@ -5,9 +5,12 @@ from ..formats.gridenvelope2d import GridEnvelope2D
 from ..formats.raster import Raster
 from ..formats.inmemoryraster import InMemoryRaster
 from ..toolbox.striplib import StripManager
+import warnings
 from math import ceil
 import numpy as np
 import os.path
+import pycrs
+import sys
 
 __author__ = "Steven B. Hoek"
     
@@ -104,35 +107,38 @@ class TileManager():
         if (rowoverlap < 0) or (coloverlap < 0):
             ValueError("Overlap cannot be negative!")
         if (self.__BRedges ^ self.__TLedges) and (rowoverlap > tileheight):
+            # Bitwise XOR gives True
             raise ValueError("Overlap cannot be greater than the tile height!")
         if (self.__BRedges and self.__TLedges) and (2 * rowoverlap > tileheight):
             raise ValueError("Top and bottom edges together cannot be greater than the tile height!")
         if (self.__BRedges ^ self.__TLedges) and (coloverlap > tilewidth):
+            # Bitwise XOR gives True
             raise ValueError("Overlap cannot be greater than the tile width!")
         if (self.__BRedges and self.__TLedges) and (2 * coloverlap > tilewidth):
             raise ValueError("Left and right edges together cannot be greater than the tile width!")
         if self.__BRedges or self.__TLedges:
             if (rowoverlap == 0) and (coloverlap == 0):
-                raise Warning("Overlap indicated as zero: no edges will be added!")
+                warnings.warn("Overlap indicated as zero: no edges will be added!")
                 self.__BRedges = False
                 self.__TLedges = False
             self.__rowoverlap = rowoverlap
             self.__coloverlap = coloverlap
         else:
             if (rowoverlap > 0) or (coloverlap > 0):
-                raise Warning("No edges selected: indicated overlap will be ignored!")
+                warnings.warn("No edges selected: indicated overlap will be ignored!")
             
         # Get the first strip ready
         self.nextstrip()
 
     def __iter__(self):
-        return TileIterator(self)
+        result = TileIterator(self)
+        return result
 
     def next(self):
         '''Returns a tile with number of rows and columns'''
         # TODO: implement the option to have row and column offsets
         result = None
-        maxrowno = (self.__nrows - 1)// self.__depth # no. of strips needed, zero-based
+        maxrowno = (self.__nrows - 1) // self.__depth # no. of strips needed, zero-based
         colsleft = self.__ncols - self.__curcol
         if colsleft > self.__width:
             # Still more tiles to be written
@@ -257,8 +263,12 @@ class TileManager():
                 self.__colbuf = None
             
             # Now that we have retrieved enough particulars, prepare the result
+            eps = 0.00001
             imr = Tile("dummy_file.ext", data, self.__datatype)
-            imr.open('w', width, height, nbands, xll, yll, self.__rg.dy, nodataval)
+            imr.open('w', width, height, nbands, xll, yll, self.__rg.cellsize, nodataval)
+            if abs(self.__rg.dy - self.__rg.dx) > eps:
+                imr.dx = self.__rg.dx
+                imr.dy = self.__rg.dy
             self.__curcol += self.__width
             self.__colidx += 1
             imr.rowidx = self.__rowidx
@@ -266,7 +276,7 @@ class TileManager():
             result = imr
         return result
     
-    def nextstrip(self):
+    def nextstrip(self, parseLine=True):
         # Initialise
         result = False
         rowbuf = None
@@ -295,9 +305,9 @@ class TileManager():
                     
             # Overlap is only counted in case of edges at the bottom and to the right
             if self.__BRedges:
-                self.__curstrip = self.__stripmanager.next(rowbuf, self.__rowoverlap)
+                self.__curstrip = self.__stripmanager.next(rowbuf, self.__rowoverlap, parseLine)
             else:
-                self.__curstrip = self.__stripmanager.next(rowbuf)
+                self.__curstrip = self.__stripmanager.next(rowbuf, 0, parseLine)
             self.__curcol = 0
             self.__colidx = -1
             self.__rowidx += 1
@@ -308,6 +318,19 @@ class TileManager():
         finally:    
             return result
 
+    def reset(self, parseLine=True):
+        # Reset the source raster
+        self.__rg.reset()
+        self.__rowidx = -1 # strip index
+        self.__curcol = 0
+        self.__curstrip = None
+        self.__rowbuf = None
+        self.__colbuf = None
+
+        # Get the first strip ready
+        self.__stripmanager.reset(parseLine)
+        self.nextstrip(parseLine)
+
     def close(self):
         self.__rg = None
         self.__initialised = False
@@ -361,13 +384,70 @@ class TileManager():
     def coloverlap(self):
         return self.__coloverlap
     
+    # The following is meant to make the tilemanager subscriptable
+    def __len__(self):
+        maxrowno = (self.__nrows - 1) // self.__depth # no. of strips needed, zero-based
+        maxcolno = self.__ncols // self.__width  # no. of columns needed, zero-based
+        result = (maxcolno + 1) * (maxrowno + 1)
+        return result
+    
+    def __getitem__(self, index):
+        # Initialise
+        result = None
+        
+        try:
+            # Check input
+            if index >= len(self):
+                raise IndexError("index %s is out of bounds" % index)
+            
+            # In principle, the tilemanager is an iterable. Determine desired state of the iteration
+            maxcolno = self.__ncols // self.__width  # no. of columns needed, zero-based
+            rowno = index // (maxcolno + 1) # zero-based
+            colno = index % (maxcolno + 1)
+    
+            # Iterate to the requested position
+            if rowno < self.__rowidx:
+                # Reset 
+                if index <= maxcolno:
+                    self.reset(True)
+                else:
+                    # Move to the appropriate row
+                    self.reset(False)
+                    for __ in range(self.__rowidx, rowno - 1): self.nextstrip(False)
+                    self.nextstrip(True)
+                
+                # Move to the right column 
+                for __ in range(colno): self.next()
+            elif rowno > self.__rowidx:
+                # Appropriate row ahead - move there
+                for __ in range(self.__rowidx, rowno - 1): self.nextstrip(False)
+                self.nextstrip(True)    
+                    
+                # Move to the right column
+                for __ in range(colno - 1): self.next()           
+            else:
+                # Tile can be extracted from current strip 
+                if self.__colidx >= colno:
+                    self.__curcol = 0
+                    self.__colidx = -1
+                
+                # One of the next tiles in same row
+                for __ in range(self.__colidx, colno - 1): self.next()  
+    
+            # return tile
+            result = self.next()
+        except Exception as e:
+            print(e)
+        finally:
+            return result
 
 class TileIterator:
     ''' Class that actally makes sure that the TileManager is iterable'''
     def __init__(self, tilemanager):
+        tilemanager.reset()
         self.__tilemanager = tilemanager
     
-    def __next__(self):
+    def __next__(self, parseLine=True):
         # Initialise
         result = None
         
@@ -380,7 +460,7 @@ class TileIterator:
             result = self.__tilemanager.next()
         else:
             # Try to see whether a new strip is available
-            if self.__tilemanager.nextstrip():
+            if self.__tilemanager.nextstrip(parseLine):
                 result = self.__tilemanager.next()
             else:    
                 raise StopIteration
@@ -441,8 +521,9 @@ class TileStitcher():
             raise ValueError("Left and right edges together cannot be greater than the tile width!")
         if option != "no-edges":
             if (rowoverlap == 0) and (coloverlap == 0):
-                raise Warning("Overlap indicated as zero: no edges will be added!")
+                warnings.warn("Overlap indicated as zero: no edges will be added!")
                 option = "no-edges"
+        else: coloverlap = rowoverlap = 0
         self.__coloverlap = coloverlap
         self.__rowoverlap = rowoverlap
         self.__nbands = 1
@@ -483,7 +564,7 @@ class TileStitcher():
         '''
         # Check whether output raster is already open
         if (outputraster.nrows != 1) and (outputraster.ncols != 1):
-            raise Warning("Output raster seems to be open already!")
+            warnings.warn("Output raster seems to be open already!")
         
         # Determine minimum and maximum rowidx etc.
         tmplist = [t["rowidx"] for t in self.__tilelist]
@@ -701,6 +782,10 @@ class TileStitcher():
                             yll = rg.yll - (self.__nrows - 1) * self.__tileheight * rg.dy 
 
                         # Open the output file, write header info and prepare for writing of the lines
+                        eps = 0.00001
+                        if abs(rg.dy - rg.dx) > eps:
+                            outputraster.dx = rg.dx
+                            outputraster.dy = rg.dy
                         outputraster.open('w', ncols, nrows, nbands, xll, yll, rg.cellsize, rg.nodatavalue)
                     
                     if self.__option == "average":   
@@ -729,3 +814,316 @@ class TileStitcher():
         print("Finished stitching the tiles together!")
         if not outputraster is None:
             outputraster.close()
+
+class VirtualTileStitcher():
+    '''
+    Class that can join tiles together virtually - also possible when they have overlap. Assumption is that the 
+    last part of the file names - before the extension - contain row and column indices which are separated by 
+    underscores or another separator. Assumption is that all tiles are of the same width and height and all have
+    the same overlaps. Writes a *.vrt file to represent the tiles in a virtual raster. See also the format doc:
+    https://gdal.org/drivers/raster/vrt.html
+    '''
+    __ncols = 1
+    __nrows = 1
+    __nbands = 1 # default
+    __tilewidth = 1
+    __tileheight = 1
+    __coloverlap = 0
+    __rowoverlap = 0
+    __tilelist = []
+    __option = "bottomright"
+    
+    def __init__(self, tilewidth, tileheight, coloverlap=0, rowoverlap=0, option="no-edges"):
+        self.__tilewidth = tilewidth
+        self.__tileheight = tileheight
+        if not option in ['average', 'topleft', 'bottomright', "no-edges", "allsides"]:
+            raise ValueError("Invalid option: %s!" % option)
+        self.__option = option
+        if (coloverlap < 0) or (rowoverlap < 0):
+            raise ValueError("Overlap cannot be negative!")
+        if (option in ['topleft', 'bottomright']) and (rowoverlap > tileheight):
+            raise ValueError("Overlap cannot be greater than the tile height!")
+        if (option in ['allsides', 'average']) and (2 * rowoverlap > tileheight):
+            raise ValueError("Top and bottom edges together cannot be greater than the tile height!")
+        if (option in ['topleft', 'bottomright']) and (coloverlap > tilewidth):
+            raise ValueError("Overlap cannot be greater than the tile width!")
+        if (option in ['allsides', 'average']) and (2 * coloverlap > tilewidth):
+            raise ValueError("Left and right edges together cannot be greater than the tile width!")
+        if option != "no-edges":
+            if (rowoverlap == 0) and (coloverlap == 0):
+                warnings.warn("Overlap indicated as zero: no edges will be added!")
+                option = "no-edges"
+        else: coloverlap = rowoverlap = 0
+        self.__coloverlap = coloverlap
+        self.__rowoverlap = rowoverlap
+        self.__nbands = 1
+    
+    def add(self, filename, separator="_"):
+        '''Method for adding to the VirtualTileStitcher a tile that should be used for the stitching. It is 
+        assumed that the filename follows a pattern. Not considering the file extensions, the filenames are 
+        expected to end with numbers indicating the place of the tile in the mesh, i.e. by means of the zero-
+        based row  and the column numbers. These numbers are separated from each other and from the rest of 
+        the filename by means of separators. The default separator is an underscore "_".
+        
+        Input variables:
+        filename - name of the image file that is storing the tile
+        separator - special token that is used to separate the figures 
+        '''
+        # Initialise
+        if not os.path.exists(filename):
+            raise ValueError("File %s does not exist!" % filename)
+        
+        # Assume that the last 2 underscores in the filename are to separate 2-digit row and column indices (zero-based)
+        basename = os.path.splitext(os.path.basename(filename))[0]
+        if basename.count(separator) < 2:
+            raise ValueError("Filename %s does not contain enough separator characters!" % filename)
+        parts = basename.split("_")
+        rowidx = int(parts[-2])
+        colidx = int(parts[-1])
+        self.__tilelist.append({"rowidx": rowidx, "colidx":colidx, "filename":filename})
+
+    def process(self, RasterClass, datatype, outputfile):
+        # Check input
+        ext = os.path.splitext(outputfile)[1]
+        if ext != '.vrt': raise ValueError("File extension .vrt expected, but %s found!" % ext)
+        
+        # Determine minimum and maximum rowidx etc.
+        tmplist = [t["rowidx"] for t in self.__tilelist]
+        minrowidx, maxrowidx = min(tmplist), max(tmplist)
+        self.__nrows = maxrowidx - minrowidx + 1
+
+        # Check that the list of tiles is complete 
+        for i in range(self.__nrows):
+            tiles = list(filter(lambda t: t["rowidx"] == i, self.__tilelist))
+            if len(tiles) < (maxrowidx - minrowidx + 1):
+                raise ValueError("Row of tiles with index %s is not complete!" % i)
+
+        # Determine minimum and maximum colidx etc.
+        tmplist = [t["colidx"] for t in self.__tilelist]
+        mincolidx, maxcolidx = min(tmplist), max(tmplist)
+        self.__ncols = maxcolidx - mincolidx + 1
+        
+        # Calculate the height and width of the virtual raster
+        factor = 0
+        if self.__option in ["topleft", "bottomright"]: factor = 1
+        elif self.__option in ["allsides", "average"]: factor = 2
+        vrtwidth  = (self.__ncols * self.__tilewidth) + (factor * self.__coloverlap)
+        vrtheight = (self.__nrows * self.__tileheight) + (factor * self.__rowoverlap)
+        
+        # Determine the projection, get the GeoTransform and determine the dataAxisToSRSAxisMapping if possible
+        tile = list(filter(lambda t: t["rowidx"] == minrowidx and t["colidx"] == mincolidx, self.__tilelist))[0]
+        fn = tile["filename"]
+        if not os.path.exists(fn): raise ValueError("File %s does not exist" % fn)    
+        
+        # Try to obtain the necessary info from the UL tile
+        nbands = 1
+        xul, yul = 0.0, 0.0
+        dx = dy = 0.1
+        rot1 = rot2 = 0.0
+        crs = None
+        try:
+            # Find out a few things about the outputfile. The tiles should all be in the same folder
+            curdir = os.path.dirname(os.path.realpath(sys.argv[0]))
+            if not os.path.isabs(outputfile):
+                outputfile = os.path.normpath(os.path.join(curdir, outputfile))
+            outputdir = os.path.realpath(os.path.dirname(outputfile))
+            
+            # Okay, the variable outputfile contains an absolute path now
+            samedir = False
+            inputdir = os.path.dirname(fn)
+            if (os.path.isabs(inputdir)):
+                if outputdir == os.path.realpath(inputdir): samedir = True
+            else:
+                inputdir = os.path.join(curdir, inputdir)
+                if outputdir == os.path.realpath(inputdir): samedir = True
+
+            # If the folder is not the same, find out which relative path we can use for the tiles
+            if not samedir:
+                pathlen_out = len(str(outputdir).split(os.path.sep))
+                inputdir = os.path.realpath(inputdir)
+                pathlen_inp = len(str(inputdir).split(os.path.sep))
+                pathlen_dif = pathlen_out - pathlen_inp
+                if (pathlen_dif > 0) and ( pathlen_dif <= 2):
+                    # We can use relative paths
+                    dirlist = str(inputdir).split(os.path.sep)
+                    reldir = os.path.join(".\\", *dirlist[pathlen_inp - pathlen_dif:])
+                else: reldir = ""
+                
+            # Open the tile and probe
+            rg = RasterClass(fn, datatype)
+            rg.open('r')
+            crs = rg.crs
+            nodatavalue = rg.nodatavalue
+            arr = np.array(rg.next())
+            datatype = str(arr.dtype).title()
+            
+            # Let's first get the coordinates of the upper left corner
+            if self.__option in ["allsides", "average"]:
+                xul, yul = rg.xll, rg.yll + (self.__tileheight + 2 * self.__rowoverlap) * rg.dy 
+            elif self.__option in ["topleft", "bottomright"]:   
+                xul, yul = rg.xll, rg.yll + (self.__tileheight + self.__rowoverlap) * rg.dy
+            else:
+                xul, yul = rg.xll, rg.yll + self.__tileheight * rg.dy
+            
+            # Check cellsize
+            dx = rg.dx
+            dy = rg.dy
+            
+            # Check the number of bands
+            if hasattr(rg, "nbands"): nbands = rg.nbands
+            print("First tile detected. It has %s bands ..." % nbands)           
+        finally:
+            rg.close()
+            
+        # Prepare to write the first few lines to the outputfile 
+        lines = []
+        lines.append(self.get_vrt_dataset(vrtwidth, vrtheight))
+        
+        if crs is None: crs = pycrs.parse.from_epsg_code(4326) # default
+        if hasattr(crs, "wkt"): crs_as_wkt = crs.wkt 
+        else: raise ValueError("Unable to establish CRS")
+        srs_str = self.get_srs(crs_as_wkt, 2, 1) # not sure about axis numbers
+        lines.append(srs_str)
+        geo_trf_str = self.get_geo_transform([xul, dx, rot1, yul, rot2, dy])
+        lines.append(geo_trf_str)
+        
+        # Loop over the bands
+        for h in range(nbands):
+            vrt_raster_band_str = self.get_vrt_raster_band(h+1, datatype)
+            lines.append(vrt_raster_band_str)
+            lines.append(self.get_nodata(nodatavalue, datatype))    
+                                      
+            # Gray, Palette, Red, Green, Blue, Alpha, Hue, Saturation, Lightness, Cyan, Magenta, Yellow, Black, or Unknown.
+            color_str = self.get_color_interp("Gray")
+            lines.append(color_str)
+            
+            # Now loop over the tiles. The output raster will be written strip by strip
+            yoff = 0
+            for i in range(minrowidx, maxrowidx + 1):
+                xoff = 0
+                for k in range(mincolidx, maxcolidx + 1):
+                    # Get hold of the right tile
+                    tile = list(filter(lambda t: t["rowidx"] == i and t["colidx"] == k, self.__tilelist))[0]
+                    filepath = os.path.normpath(os.path.join(curdir, tile["filename"]))
+                    if not os.path.exists(filepath): raise ValueError("File %s does not exist!" % filepath)
+                    if samedir: 
+                        fn = os.path.basename(tile["filename"])
+                        relative = True
+                    else:
+                        if reldir != "": 
+                            fn = os.path.join(reldir, tile["filename"])
+                            relative = True
+                        else:
+                            fn = os.path.realpath(os.path.join(curdir, tile["filename"]))
+                            relative = False
+
+                    # Get the lines necessary for including this tile                       
+                    src_filename = self.get_src_filename(relative, fn)
+                    
+                    # We could check here whether the number of bands in the tiles is really the same, 
+                    # but it's unlikely to be otherwise, so for the meantime we asumme it's okay
+                    rg = RasterClass(filepath, datatype)
+                    rg.open('r')
+                    src_nodata = self.get_src_nodata(rg.nodatavalue, datatype)
+                    rg.close()
+                    rstsizex   = self.__tilewidth + (factor * self.__coloverlap)
+                    rstsizey   = self.__tileheight + (factor * self.__rowoverlap)
+                    src_props  = self.get_src_props(rstsizex , rstsizey, datatype, 0, 0) # TODO: blksizey???
+                    src_rect   = self.get_src_rect(0, 0, rstsizex, rstsizey)
+                    dst_rect   = self.get_dst_rect(xoff, yoff, rstsizex, rstsizey)
+                    
+                    # Okay, let's use the above parameters now to compose the XML for 1 tile
+                    src_str         = self.get_src_str(src_filename, src_props, src_rect, dst_rect, src_nodata)
+                    complex_src_str = self.get_complex_src(src_str)
+                    lines.append(complex_src_str + "\n")
+                    
+                    # Prepare for the next tile
+                    xoff += self.__tilewidth
+                
+                # New row
+                yoff += self.__tileheight
+                            
+            # Write end </VRTRasterBand>
+            lines.append("    </VRTRasterBand>\n")
+            
+        # Write end </VRTDataset>
+        lines.append("</VRTDataset>\n")
+        
+        # Now write the lines to file
+        with open(outputfile, 'wt') as f:
+            f.writelines(lines)
+        print("File %s written to disk." % outputfile)
+        
+    
+    '''With the following functions we generate the necessary XML'''
+    def get_vrt_dataset(self, width, height):
+        if width <= 0 or height <= 0: raise ValueError("Invalid dimensions!")
+        result = '<VRTDataset rasterXSize="%s" rasterYSize="%s">\n' % (width, height) 
+        return result
+    
+    def get_srs(self, srs_wkt, axis_no1 = 2, axis_no2 = 1):
+        result = '  <SRS dataAxisToSRSAxisMapping="%s,%s">%s</SRS>\n' % (axis_no1, axis_no2, srs_wkt)
+        return result
+    
+    def get_geo_transform(self, trfvals):
+        # we need six value which indicate the affine geotransformation - see:
+        # https://homepages.inf.ed.ac.uk/rbf/HIPR2/affine.htm
+        if len(trfvals) != 6: raise ValueError("Invalid input values found for affine transformation!")
+        [xul, dx, rot1, yul, rot2, dy] = trfvals
+        result = '  <GeoTransform>%s, %s, %s, %s, %s, %s</GeoTransform>\n' % (xul, dx, rot1, yul, rot2, -1*dy)
+        return result
+        
+    def get_vrt_raster_band(self, bandno, datatype):
+        result = '  <VRTRasterBand dataType="%s" band="%s">\n' % (datatype, bandno)
+        return result
+    
+    def get_nodata(self, nodatavalue, datatype):
+        if datatype[0].lower() == 'i': nodatavalue = int(nodatavalue)
+        result = '    <NoDataValue>%s</NoDataValue>\n' % nodatavalue
+        return result
+    
+    def get_color_interp(self, clrstr):
+        result = '    <ColorInterp>%s</ColorInterp>\n' % clrstr
+        return result
+
+    def get_complex_src(self, srcstr):
+        # Get the XML describing the source of this tile
+        result = '    <ComplexSource resampling="nearest">\n%s\n    </ComplexSource>' % srcstr
+        return result
+    
+    def get_src_str(self, src_filename, src_props, src_rect, dst_rect, src_nodata):
+        # Concatenate all pieces of the source string together
+        result = src_filename + "\n" + self.get_src_band() + "\n" + src_props
+        result += src_rect + "\n" + dst_rect + "\n" + src_nodata
+        return result
+    
+    def get_src_filename(self, relative, filepath):
+        # Get the source tag for this tile
+        reltovrt = 1 if relative else 0
+        result = '      <SourceFilename relativeToVRT="%s">%s</SourceFilename>' % (reltovrt, filepath)
+        return result
+    
+    def get_src_band(self):
+        result = '      <SourceBand>1</SourceBand>'
+        return result
+    
+    def get_src_props(self, rstsizex, rstsizey, datatype, blksizex=0, blksizey=0):
+        if blksizex == 0: blksizex = rstsizex
+        # if blksizey == 0: blksizey = rstsizey
+        # BlockYSize="%s" - it's not clear how the value for this should be calculated but it's optional
+        result = '<SourceProperties RasterXSize="%s" RasterYSize="%s" BlockXSize="%s" DataType="%s" />\n'        
+        result = (6 * ' ') + result % (rstsizex, rstsizey, blksizex, datatype)
+        return result
+    
+    def get_src_rect(self, xoff, yoff, xsize, ysize):
+        result = '      <SrcRect xOff="%s" yOff="%s" xSize="%s" ySize="%s" />' % (xoff, yoff, xsize, ysize)
+        return result
+    
+    def get_dst_rect(self, xoff, yoff, xsize, ysize):
+        result = '      <DstRect xOff="%s" yOff="%s" xSize="%s" ySize="%s" />' % (xoff, yoff, xsize, ysize)
+        return result
+    
+    def get_src_nodata(self, nodatavalue, datatype):
+        if datatype[0].lower() == 'i': nodatavalue = int(nodatavalue)
+        result = '      <NODATA>%s</NODATA>' % nodatavalue
+        return result
